
#-------------------------------------------------------------------------------
# To DO: Add documentation of same dimensions exp and exp_cor
# To DO: Add documentation for units
# TO DO: Write Documentation

# library(s2dv) # InsertDim, ...
# library(multiApply) # Apply
# library(ClimProjDiags) # Subset
# library(zeallot) # multiple assignment %<-%
# library(SPEI) # estimation of evapotranspiration (e.g. hargreaves function)
# library(zoo) # rollapply, as.Date
# library(TLMoments) # pwd (Calculate probability weighted moments based on fit)
# library(lmomco) # pwm2lmom (probability weighted moments to L-Moments)
# library(lubridate) # year(), month(),...

# library(CSTools)
#-------------------------------------------------------------------------------
#'@param exp A list with the 's2dv_cube' object class containing the seasonal 
#'  forecast experiment in the data element.
#'@param exp_cor An object of class \code{s2dv_cube} in which the quantile 
#'  PeriodSPEI should be applied. If it is not specified, the PeriodSPEI
#'  is calculated from object 'exp'.
#'@param pet Multidimensional array containing the Potential EvapoTranspiration 
#'  data. If it is NULL it is calculated using pet_method. It is NULL by default.
#'@param time_dim
#'@param leadtime_dim
#'@param memb_dim
#'@param lat_dim
#--- other params from CSIndicators
#'@param accum
#'@param start
#'@param end
#' 
# --- other params
#'@param pet_method
#'@param standardization
#'@param params 
#'@param param_error 
#'@param handle_infinity
#'@param cross_validation
#'@param method
#'@param distribution
#'@param fit
#'@param ncores
CST_PeriodSPEI <- function(exp, exp_cor = NULL, pet = NULL, 
                           time_dim = 'syear', leadtime_dim = 'time', 
                           memb_dim =  'ensemble', lat_dim = 'latitude',
                           accum = 1, start = NULL, end = NULL,
                           pet_method = c('hargreaves', 'hargreaves'),
                           standardization = TRUE,
                           params = NULL, # not used if standardization is FALSE
                           param_error = -9999, # not used if standardization is FALSE
                           handle_infinity = FALSE, # not used if standardization is FALSE
                           cross_validation = FALSE, # not used if standardization is FALSE
                           method = 'parametric', distribution = 'log-Logistic', 
                           fit = 'ub-pwm', # not used if standardization is FALSE
                           ncores = 4) {
                          
  #-----------------------------------------------------------------------------                     
  # Part (1): Initial structure checks
  # coordinates
  .KnownLatNames <- CSTools:::.KnownLatNames()

  if (!any(names(exp[[1]]$coords) %in% .KnownLatNames)) {
    stop("Spatial coordinate names of parameter 'obsL' do not match any ",
         "of the names accepted by the package.")
  }
  lat_name <- names(exp[[1]]$coords)[[which(names(exp[[1]]$coords) %in% .KnownLatNames)]]

  #-----------------------------------------------------------------------------
  # Part (2): Call to PeriodSPEI

  res <- PeriodSPEI(exp = lapply(exp, function(x) x$data), 
                    dates = exp[[1]]$attrs$Dates, 
                    lat = exp[[1]]$coords[[lat_name]], 
                    exp_cor = lapply(exp_cor, function(x) x$data), 
                    pet = pet, 
                    time_dim = time_dim, leadtime_dim = leadtime_dim, 
                    memb_dim = memb_dim, lat_dim = lat_dim,
                    accum = accum, start = start, end = end, 
                    pet_method = pet_method, standardization = standardization,
                    params = params, param_error = param_error, 
                    handle_infinity = handle_infinity, 
                    cross_validation = cross_validation,
                    method = method, distribution = distribution, 
                    fit = fit, ncores = ncores)
  return(res)

}



PeriodSPEI <- function(exp, dates, lat, 
                       exp_cor = NULL, pet = NULL, 
                       time_dim = 'syear', leadtime_dim = 'time', 
                       memb_dim = 'ensemble', lat_dim = 'latitude',
                       accum = 1, start = NULL, end = NULL, 
                       pet_method = c('hargreaves', 'hargreaves'), 
                       standardization = TRUE,
                       params = NULL, param_error = -9999, 
                       handle_infinity = FALSE, cross_validation = FALSE,
                       method = 'parametric', distribution = 'log-Logistic', 
                       fit = 'ub-pwm', ncores = 4) {

  #-----------------------------------------------------------------------------

  # Part (1): Initial checks

  ## pet and pet_method
  if (is.null(pet_method) & is.null(pet)) {
    warning("Parameter 'pet_method' cannot be NULL if pet is not provided.")
  }
  if (!is.null(exp_cor)) {
    if (length(pet_method) == 1) {
      pet_method <- rep(pet_method, 2)
    }
    if (length(exp_cor) < 1) {
      exp_cor <- NULL
    }
  } else {
    if (lenght(pet_method) > 1) {
      warning("Parameter 'pet_method' is of length 2, only first value will be used.")
      pet_method <- pet_method[1]
    }
  }
  ## exp 
  ## TO DO: Check order: tasmax, tasmin, prlr
  if (all(c('tasmin', 'tasmax', 'prlr') %in% names(exp))) {
    # hargreaves modified: 'tasmin', 'tasmax', 'prlr' and 'lat'
    if (!any(pet_method %in% c('hargreaves_modified', 'hargreaves'))) {
      warning("Parameter 'pet_method' needs to be 'hargreaves_modified' or 'hargreaves'.")
      pet_method[1] <- 'hargreaves_modified'
    }
  } else if (all(c('tasmin', 'tasmax') %in% names(exp))) {
    # hargreaves: 'tasmin', 'tasmax' and 'lat'
    if (pet_method != c('hargreaves')) {
      warning("Parameter 'pet_method' needs to be 'hargreaves'.")
      pet_method[1] <- 'hargreaves'
    }
  
  } else if ('tas' %in% names(exp)) {
    # thornthwaite: 'tas' (mean), 'lat'
    pet_method[1] <- 'thornthwaite'

  } else {
    stop("Parameter 'exp' needs to be a list with the needed variables.")
  }

  #  warning(' needs to be in C and precipitation in mm/month') # there is no check

  # check if accumulation period is possible
  if (accum > dim(exp[[1]])[leadtime_dim]) {
    stop(paste0('ERROR: Cannot compute accumulation of ', accum, ' months because loaded data has only ', 
         dim(exp[[1]])[leadtime_dim], ' months.'))
  }
  # method
  if (!(method %in% c('parametric', 'non-parametric'))) {
    stop('pcX script error: SPEI can be only computed using the following approach: parametric or non-parametric')
  }
  # distribution
  if (!(distribution %in% c('log-Logistic', 'Gamma', 'PearsonIII'))) {
    stop('Distrib must be one of "log-Logistic", "Gamma" or "PearsonIII"')
  }
  # fit
  if (!(fit %in% c('max-lik', 'ub-pwm', 'pp-pwm'))) {
    stop('Method must be one of "ub-pwm" (default), "pp-pwm" or "max-lik"')
  }

  #-----------------------------------------------------------------------------
  # Part (2): preparation
  
  # complete dates
  ini_date <- as.Date(paste(lubridate::year(min(dates)), 01, 01, sep= '-'))
  end_date <- as.Date(paste(lubridate::year(max(dates)), 12, 31, sep='-'))
  dates_complete_daily <- as.Date(ini_date:end_date)

  dates_complete_daily_to_monthly <- lubridate::day(dates_complete_daily)
  dates_complete_monthly <- dates_complete_daily[which(dates_complete_daily_to_monthly == 1)]
  dates_monthly <- array(0, dim=length(dates_complete_monthly))
  for (dd in 1:length(dates)) {
    ii <- which(dates_complete_monthly == as.Date(paste(lubridate::year(dates[dd]), lubridate::month(dates[dd]), 01, sep = '-')))
    dates_monthly[ii] <- 1
  }

  #-----------------------------------------------------------------------------
  # Part (3): Compute PeriodSPEI
  k = 0
  spei_res <- NULL
  computed_pet <- FALSE

  for (data in .return2list(exp, exp_cor)) {
    k = k + 1
    # Evapotranspiration estimation (unless pet is already provided)
    if (is.null(pet) | computed_pet) {
      pet <- evapotranspiration(data, dates_monthly, pet_method = pet_method[k], time_dim, 
                                leadtime_dim, memb_dim, lat_dim, ncores)
      computed_pet <- TRUE
    }
    # Accumulation
    diff_P_PET <- data$prlr - pet
    data_accum <- accumulation(diff_P_PET, dates_monthly, accum, time_dim, 
                               leadtime_dim, memb_dim, ncores)
    # Standardization: 
    if (standardization == TRUE) {
      spei_dat <- spei_standardization(data_accum = data_accum,
                                      leadtime_dim = leadtime_dim,
                                      time_dim = time_dim,
                                      memb_dim = memb_dim,
                                      cross_validation = cross_validation,
                                      handle_infinity = handle_infinity,
                                      ncores = ncores,
                                      accum = accum,
                                      param_error = param_error,
                                      params = params, method = method, 
                                      distribution = distribution, fit = fit)
      params <- spei_dat$params
    } else {
      spei_dat <- data_accum
    }
    spei_res[[k]] <- spei_dat[[1]]
  }

  if (standardization) {
    spei_res[[k+1]] <- params
    if (is.null(exp_cor)) {
      names(spei_res) <- c('exp', 'params')
    } else {
      names(spei_res) <- c('exp', 'exp_cor', 'params')
    }
  } else {
    if (is.null(exp_cor)) {
      names(spei_res) <- c('exp')
    } else {
      names(spei_res) <- c('exp', 'exp_cor')
    }
  }

  return(spei_res)

}


.return2list <- function(data1, data2 = NULL) {
  if (is.null(data2)) {
    return(list(data1))
  } else {
    return(list(data1, data2))
  }
}




evapotranspiration <- function(data, dates_monthly, pet_method = 'hargreaves', 
                                time_dim = 'syear', leadtime_dim = 'time', memb_dim = 'ensemble', 
                                lat_dim = 'latitude', 
                                ncores = 4) {

  ## corrected::
  lat_mask <- array(lat, dim = c(1, length(lat)))
  names(dim(lat_mask)) <- c('dat', lat_dim)

  # extract mask of NA locations to return to NA the final result

  mask_NA <- array(1, dim = dim(data[[1]]))
  if (pet_method == 'hargreaves') {
    mask_NA[which(is.na(data$tasmax))] <- 0
    mask_NA[which(is.na(data$tasmin))] <- 0
  } else if (pet_method == 'hargreaves_modified') {
    mask_NA[which(is.na(data$tasmax))] <- 0
    mask_NA[which(is.na(data$tasmin))] <- 0
    mask_NA[which(is.na(data$prlr))] <- 0
  } else if (pet_method == 'thornthwaite') {
    mask_NA[which(is.na(data$tas))] <- 0
  }

  # replace NA with 0
  for (dd in 1:length(data)) {
    data[[dd]][which(is.na(data[[dd]]))] <- 0
  }

  # prepare data
  target_dims_data <- lapply(data, function(x) rep(c(leadtime_dim, time_dim), 1))
  # file <- tempfile()
  # file = 'out.txt'
  # sink(file)
  PET_estimated <- Apply(data = c(list(lat_mask = lat_mask), data), 
                         target_dims = c(list(lat_mask = 'dat'), target_dims_data), 
                         output_dims = c(leadtime_dim, time_dim),
                         pet_method = pet_method, 
                         dates_monthly = dates_monthly,
                         leadtime_dim = leadtime_dim, time_dim = time_dim,
                         fun = atomic_pet, ncores = ncores)$output1
  # sink(NULL)
  # captured_message <- readLines(file)
  # print(paste0('Length captured!!', length(captured_message)))
  # print(paste0('unique(captured_message)', length(unique(captured_message))))

  # reorder dims in PET_estimated
  pos <- match(names(dim(data[[1]])),names(dim(PET_estimated)))
  PET_estimated <- aperm(PET_estimated, pos)

  # restore original NAs from mask_NA
  PET_estimated[which(mask_NA == 0 )] <- NA

  return(PET_estimated)

}

accumulation <- function(diff_P_PET, dates_monthly, accum = 1, 
                         time_dim = 'syear', leadtime_dim = 'time', 
                         memb_dim = 'ensemble', ncores = 4) {
  
  if (!time_dim %in% names(dim(diff_P_PET))) {
    diff_P_PET <- InsertDim(diff_P_PET, posdim = 1, lendim = 1, name = time_dim)
  }
  if (!leadtime_dim %in% names(dim(diff_P_PET))) {
    diff_P_PET <- InsertDim(diff_P_PET, posdim = 1, lendim = 1, name = leadtime_dim)
  }
  
  accum_result <- Apply(data = list(diff_P_PET), 
                        target_dims = list(diff_P_PET = c(leadtime_dim,time_dim)), 
                        dates_monthly = dates_monthly,
                        accum = accum,
                        output_dims = c(leadtime_dim, time_dim), #c('time','sdate'),
                        leadtime_dim = leadtime_dim, time_dim = time_dim,
                        fun = atomic_accum, ncores = ncores)
    
  # recover essential lost dims (if they had length 1 they'd have been dropped in previous step):
  for (d in c(time_dim, leadtime_dim, 'latitude', 'longitude', memb_dim)) {
    if(!d %in% names(dim(accum_result$output1))){ 
      accum_result$output1 <- InsertDim(data = accum_result$output1, posdim = length(names(dim(accum_result$output1))) + 1, lendim = 1, name = d)
    }
  }
  
  # reorder dims in accum_result
    dims_order <- array(NA, length(dim(diff_P_PET)))
    for (ord in 1:length(dim(diff_P_PET))){
      dims_order[ord] <- which(names(dim(accum_result$output1)) == names(dim(diff_P_PET))[ord])
    }
    data_accum <- aperm(accum_result$output1, dims_order)

  return(data_accum)

}

spei_standardization <- function(data_accum,
                                 leadtime_dim, time_dim, memb_dim, handle_infinity, 
                                 cross_validation, ncores, accum, param_error,
                                 params = NULL, method ='parametric', 
                                 distribution = 'log-Logistic', fit='ub-pwm') {

  n_leadtimes <- dim(data_accum)[leadtime_dim]
  n_sdates_params <- dim(data_accum)[time_dim]

  if (!cross_validation) {
    n_sdates_params <- 1
  }

  if (is.null(params)) {
    params <- array(NA, dim = c(n_sdates_params, n_leadtimes, coef = 3))
    names(dim(params)) <- c(time_dim, leadtime_dim, 'coef')
  } else if (length(dim(params)) < 2) {
    params <- array(params, dim = c(length(params), n_sdates_params, n_leadtimes))
    params <- aperm(params, c(2,3,1)) # dim(params): [time_dim, leadtime_dim, coef] with the values repeated each time_dim and leadtime_dim
    names(dim(params)) <- c(time_dim, leadtime_dim, 'coef')
  }

  spei <- Apply(data = list(data = data_accum, params = params), 
                target_dims = list(data = c(leadtime_dim, time_dim, memb_dim), 
                                   params = c(time_dim, leadtime_dim, 'coef')),
                output_dims = list(data_spei = c(leadtime_dim, time_dim, memb_dim), 
                                   params = c(time_dim, leadtime_dim, 'coef')), 
                leadtime_dim = leadtime_dim, 
                time_dim = time_dim,
                dim_memb = memb_dim,
                handle_infinity = handle_infinity,
                cross_validation = cross_validation,
                method = method, distribution = distribution, fit = fit, 
                ref_period = NULL,
                param_error = param_error,
                fun = atomic_spei, 
                ncores = ncores)

  return(spei) # spei is a list of data_spei and params
}

atomic_pet <- function(lat_mask = NULL, data2 = NULL, data3 = NULL, data4 = NULL, 
                       pet_method, dates_monthly, leadtime_dim, time_dim) {
 
  dims <- dim(data2)

  # create a vector from data but adding 0 to achive complete time series 
  # of the considered period 
  # (starting in January of the first year) so that the solar radiation 
  # estimation is computed in each case for the correct month

  if (!is.null(data2)) {
    data_tmp <- as.vector(data2)
    data2 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data2[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  if (!is.null(data3)) {
    data_tmp <- as.vector(data3)
    data3 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data3[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  if (!is.null(data4)) {
    data_tmp <- as.vector(data4)
    data4 <- array(0, dim = length(dates_monthly))
    count <- 1
    for (dd in 1:length(dates_monthly)) {
      if (dates_monthly[dd] == 1) {
        data4[dd] <- data_tmp[count]
        count <- count + 1
      }
    }
    rm(data_tmp)
  }
  if (pet_method == 'hargreaves') {
    # NOTE EVA: if NA.RM is FALSE this gives error
    x_PET <- hargreaves(Tmin = as.vector(data3), Tmax = as.vector(data2), 
                        lat = lat_mask, na.rm = TRUE)
    # line to return the vector to the size of the actual original data
    x_PET <- x_PET[which(dates_monthly == 1)]
    PET <- array(x_PET, dim = dims)
  }

  if (pet_method == 'hargreaves_modified') {
    x_PET <- hargreaves(Tmin = as.vector(data3), Tmax = as.vector(data2), 
                        lat = lat_mask, Pre = as.vector(data4), na.rm = TRUE)
    x_PET <- x_PET[which(dates_monthly == 1)]  
    PET <- array(x_PET, dim = dims)
  }
  
  if (pet_method == 'thornthwaite') {
    x_PET <- thornthwaite(as.vector(data2), lat = lat_mask, na.rm = TRUE)
    # line to return the vector to the size of the actual original data
    x_PET <- x_PET[which(dates_monthly == 1)] 
    PET <- array(x_PET, dim = dims)
  }
  return(PET)
}

atomic_accum <- function(data, dates_monthly, accum, leadtime_dim, time_dim) {

  # data:[time, syear]
  ftime <- dim(data)[1]
  n_sdates <- dim(data)[2]
  
  data_vector <- array(0, dim = length(dates_monthly))
  count <- 1
  for (dd in 1:length(dates_monthly)) {
    if (dates_monthly[dd] == 1){
      data_vector[dd] <- as.vector(data)[count]
      count <- count + 1
    }
  }

  # Accumulation at different timescales
  data_sum_x <- rollapply(data_vector, accum, sum) # rollapply {zoo} A generic function for applying a function to rolling margins of an array.
  data_sum_x <- c(rep(NA, accum-1), data_sum_x) # adds as many NAs as needed at the begining to account for the months that cannot be added
                                                # (depends on accu) and so that the position in the vector corresponds to the accumulated 
                                                # of the previous months (instead of the accumulated of the next months)
  data_sum_x <- data_sum_x[which(dates_monthly == 1)] # discard the months that don't appear in the original data
  accum_result <- array(data_sum_x, dim = c(ftime,n_sdates)) # return to matrix form
  if (accum > 1) {
    accum_result[1:(accum-1),] <- NA # replace by NA when the accumulation corresponds to months that where not present in the original data
  }

  return(accum_result)
  
}
atomic_spei <- function(data, params, leadtime_dim, time_dim, dim_memb, ref_period, 
                        handle_infinity, cross_validation, param_error, method = 'parametric', 
                        distribution = 'log-Logistic', fit = 'ub-pwm') {
  # data: [time, sdate, memb]
  if (is.null(ref_period)) {
    ref.start <- NULL
    ref.end <- NULL
  } else {
    ref.start <- ref_period[[1]]
    ref.end <- ref_period[[2]]
  }

  if (all(is.na(data))) {
    speiX <- array(NA, dim(data))
  } else if (var(data, na.rm=T) == 0) { # if the data [time, sdate, memb] has no variability it will raise an error further down, 
                                        # so we assign a value to the result and skip the step
    speiX <- array(param_error, dim(data))
  } else {
    c(speiX, params) %<-% spX(data, ref.start=ref.start, ref.end=ref.end, params = params, 
                              method = method, distribution = distribution, fit = fit, 
                              leadtime_dim = leadtime_dim, time_dim = time_dim, dim_memb = dim_memb, 
                              handle_infinity = handle_infinity, cross_validation = cross_validation)
  }

  return(list(speiX, params))
  
}

#-------------------------------------------------------------------------------

### Support functions

spX <- function(acc_data, ref.start, ref.end, params = params, na.rm = TRUE, 
                method = 'parametric', distribution = 'log-Logistic', fit = 'ub-pwm', 
                leadtime_dim, time_dim, dim_memb, handle_infinity, cross_validation){

  # acc_data: [ftime, syear, nmemb]
  # params: [syear, ftime, coef]

  n_coef_max <- dim(params)['coef'] # maximum number of parameters needed to define any of the considered distributions

  ftime <- as.numeric(dim(acc_data)[leadtime_dim])
  n_sdates <- as.numeric(dim(acc_data)[time_dim])
  nmemb <- as.numeric(dim(acc_data)[dim_memb])

  spei_mod <- array(data = NA, dim = c(ftime, n_sdates, nmemb))
  names(dim(spei_mod)) <- c(leadtime_dim, time_dim, dim_memb)
  if (cross_validation) {
    params_result <- array(data = NA, dim = c(n_sdates, ftime, n_coef_max))
  } else {
    params_result <- array(data = NA, dim = c(1, ftime, n_coef_max))
  }

  for (ff in 1:ftime) { # treat months independently  
    Accum_D_temp <- ClimProjDiags::Subset(acc_data, along = leadtime_dim, indices = ff, drop = FALSE)
    params_tmp <- if (all(is.na(params))) {NULL} else {params[,ff,]}
    c(x_spei_mod, params_ff) %<-% spX_ftime(Accum_D_temp, na.rm = TRUE, method = method, distribution = distribution, 
                                            fit = fit, ref.start = ref.start, ref.end = ref.end, 
                                            params = params_tmp, n_sdates = n_sdates, nmemb = nmemb, 
                                            handle_infinity = handle_infinity, cross_validation = cross_validation)
    spei_mod[ff,,] <- x_spei_mod
    coef_names <- names(params_ff)
    if (length(params_ff) < n_coef_max){ # lengthen dimension coef of params_ff in case it doesn't match the corresponding dimension of parms_months
      params_ff <- append(params_ff, array(NA, dim = n_coef_max - length(params_ff)))
      coef_names <- append(coef_names, '')
    }
    params_result[,ff,] <- params_ff
  }
  
  colnames(params_result) <- coef_names
  names(dim(params_result)) <- c(time_dim, leadtime_dim, 'coef')

  return(list(spei = spei_mod, params = params_result))
}

#-------------------------------------------------------------------------------

# HandleInfinity:
spX_ftime <- function(data, distribution = 'log-Logistic', fit = 'ub-pwm', na.rm = TRUE, method = 'parametric', n_sdates, nmemb,
                             ref.start = NULL, ref.end = NULL, params = NULL, handle_infinity = FALSE, cross_validation = FALSE) {
  
  # data:[ftime = 1, syear, ensemble]
  
  if (method == 'non-parametric') { 
    
    if (anyNA(data) && na.rm == FALSE) {
      stop('Error: Data must not contain NAs')
    }
    
    bp = matrix(0, length(data), 1)
    for (i in 1:length(data)) {
      bp[i,1] = sum(data[] <= data[i]);  # Writes the rank of the data
    }
    
    SPEI = qnorm((bp-0.44)/(length(data)+0.12))
    
    return(SPEI) # it won't return params to be used in exp_cor; also it is not using handle_infinity nor cross_validation
    
  } else {

    std_index <- array(NA, c(n_sdates, nmemb))
  
    if (anyNA(data) && na.rm == FALSE) {
      stop('Error: Data must not contain NAs')
    }

    
    coef = switch(distribution,
                  "Gamma" = array(NA, dim = 2, dimnames = list(c('alpha','beta'))),
                  "log-Logistic" = array(NA, dim = 3, dimnames = list(c('xi','alpha','kappa'))),
                  "PearsonIII" = array(NA, dim = 3, dimnames = list(c('mu','sigma','gamma'))))

    dim_one <- length(coef)

    if (!is.null(params)) {
      if (length(params) != dim_one) {
        stop(paste0('parameters array should have dimensions [', dim_one, ']'))
      }
    }
    
    # Select window if necessary
    if (!is.null(ref.start) && !is.null(ref.end)) {
      data.fit <- window(data,ref.start,ref.end)	
    } else {
      data.fit <- data
    }

    if (cross_validation == "TRUE") {  
      loop_years <- n_sdates
    } else {
      loop_years <- 1
    }

    params_result <- array(NA, dim = c(loop_years, dim_one))
    colnames(params_result) <- names(coef)
    
    for (nsd in 1:loop_years) { # loop over years (in case of cross_validation)
      # Cumulative series (acu)
      if (cross_validation == TRUE){
        acu <- as.vector(data.fit[,-nsd,])
      } else {
        acu <- as.vector(data.fit)
      }
        
      acu.sorted <- sort.default(acu, method = "quick")
      acu.sorted <- acu.sorted[!is.na(acu.sorted)] # remove NAs (no need if(na.rm) because if there are NA and na.rm=F we don't get to this point)
      if (length(acu.sorted)!=0){ # else all acu was NA and we don't need to continue with this case     
        acu_sd = sd(acu.sorted)
        if (!is.na(acu_sd)){
        if (acu_sd != 0){
          if(distribution != "log-Logistic") {
            pze <- sum(acu==0)/length(acu)
            acu.sorted = acu.sorted[acu.sorted > 0]
          }
          if (!is.null(params)) {
            f_params = as.vector(params)
            params_result[nsd,] <- f_params
          } else {
            if (length(acu.sorted) >= 4){ # else coef will be NA
              # Calculate probability weighted moments based on fit with lmomco or TLMoments
              pwm = switch(fit,
                            "pp-pwm" = pwm.pp(acu.sorted,-0.35,0, nmom=3),
                            pwm.ub(acu.sorted, nmom=3)
                            #TLMoments::PWM(acu.sorted, order=0:2)
              )

              # Check L-moments validity
              lmom <- pwm2lmom(pwm)          
              if (!(!are.lmom.valid(lmom) || anyNA(lmom[[1]]) || any(is.nan(lmom[[1]])))){
                
                # lmom fortran functions need specific inputs L1, L2, T3
                # this is handled by lmomco internally with lmorph
                fortran_vec = c(lmom$lambdas[1:2], lmom$ratios[3])

                # Calculate parameters based on distribution with lmom then lmomco
                f_params = switch(distribution,
                          "log-Logistic" = tryCatch(lmom::pelglo(fortran_vec), error = function(e){ parglo(lmom)$para }),
                          "Gamma" = tryCatch(lmom::pelgam(fortran_vec), error = function(e){ pargam(lmom)$para }),
                          "PearsonIII" = tryCatch(lmom::pelpe3(fortran_vec), error = function(e){ parpe3(lmom)$para })
                )

                # Adjust if user chose log-Logistic and max-lik
                if(distribution == 'log-Logistic' && fit == 'max-lik'){
                  f_params = parglo.maxlik(acu.sorted, f_params)$para
                }
                params_result[nsd,] <- f_params
              } # end if dor the case the L-moments are not valid (std_index will be NA)
            } # end if for the case there are not enough values to estimate the parameters (std_index will be NA)
          } # end estimation of f_param
          # Calculate cdf based on distribution with lmom
          if (all(is.na(params_result[nsd,]))){
            cdf_res <- NA
          } else {
            f_params <- params_result[nsd,]
            f_params <- f_params[which(!is.na(f_params))]
            cdf_res = switch(distribution,
                            "log-Logistic" = lmom::cdfglo(data, f_params),
                            "Gamma" = lmom::cdfgam(data, f_params),
                            "PearsonIII" = lmom::cdfpe3(data, f_params)				  				
            )
          }

          std_index_cv <- array(qnorm(cdf_res), dim = c(n_sdates,nmemb))
  
          # Adjust if user chose Gamma or PearsonIII - Not tested: For future development
          #if(distribution != 'log-Logistic'){ 
          #  std_index[ff,s] = qnorm(pze + (1-pze)*pnorm(std_index[ff,s])) # ff doesn't exist at this point
          #}
          if (cross_validation == TRUE){
            std_index[nsd,] <- std_index_cv[nsd,]  
          } else {
            std_index <- std_index_cv
          }
        }
        } # end if for the case there is no variability        
      } # end if for the case all NA in acu
    } # next year (in case of cross_validation or all done if cross_validation == F)

    if(handle_infinity == 'TRUE'){ # could also use "param_error" ?; we are giving it the min/max value of the grid point
      std_index[is.infinite(std_index) & std_index < 0] <- min(std_index[!is.infinite(std_index)])
      std_index[is.infinite(std_index) & std_index > 0] <- max(std_index[!is.infinite(std_index)]) 
    }

    return(list(std_index = std_index, params = params_result)) # f_params will be params only if cross_validation is FALSE
                                                                # (otherwise there will be one f_params per year;
                                                                # but the output params will be read only in the case that
                                                                # it is called with cross_validation FALSE)
  
  } # end parametric

}
